import time
import threading
import sys
import warnings
from pprint import pprint
import tqdm
from typing import List, Optional, overload, Union

import numpy as np
from matplotlib import pyplot as plt
from scipy.fft import fft2, ifft2, rfft2, irfft2, set_global_backend
import pyfftw

from michael960lib.math import fourier
from michael960lib.common import overrides, experimental, deprecated
from michael960lib.common import ModifyingReadOnlyObjectError, IllegalActionError

from torusgrid.grids import RealGrid2D, ComplexGrid2D, RealGrid1D, ComplexGrid1D, StateFunction
from ..weights import ParametrizedWeight
from ..sampling import DistributionSampling


class KuramotoBase:
    def __init__(self, K: float,
        theta: Union[np.ndarray, DistributionSampling], 
        omega: Union[np.ndarray, DistributionSampling], 
        weight: Union[np.ndarray, ParametrizedWeight]):

        self.K = K
        
        self._thetaobj = None
        self.theta_params = dict()
        if isinstance(theta, np.ndarray):
            self.theta = theta
        elif isinstance(theta, DistributionSampling):
            self.theta = theta.get_val()
            self.theta_params = theta.get_params()
            self._thetaobj = theta
        else:
            raise ValueError('invalid theta')

        self._omegaobj = None
        self.omega_params = dict()
        if isinstance(omega, np.ndarray):
            self.omega = omega
        elif isinstance(omega, DistributionSampling):
            self.omega = omega.get_val()
            self.omega_params = omega.get_params()
            self._omegaobj = omega
        else:
            raise ValueError('invalid omega')

        self._weightobj = None
        self.weight_params = dict()
        if isinstance(weight, np.ndarray):
            self.weight = weight
        elif isinstance(weight, ParametrizedWeight):
            self.weight = weight.get_weight()
            self.weight_params = weight.get_params()
            self._weightobj = weight
        else:
            raise ValueError('invalid weight')

        try:
            assert self.theta.shape == self.omega.shape
            assert self.theta.shape == self.weight.shape
        except AssertionError:
            raise ValueError(
            f'theta, omega, and weight have incompatible shapes: {theta.shape}, {omega.shape}, and {weight.shape}')

        self.shape = self.theta.shape
        self.rank = len(self.shape)

    def get_theta_label(self):
        if self._thetaobj is None:
            return 'custom'
        else:
            return self._thetaobj.get_label()

    def get_omega_label(self):
        if self._thetaobj is None:
            return 'custom'
        else:
            return self._omegaobj.get_label()

    def get_weight_label(self):
        if self._weightobj is None:
            return 'custom'
        else:
            return self._weightobj.get_label()



        

class KuramotoStateFunction(StateFunction):
    @overload
    def __init__(self, theta: np.ndarray): ...
    @overload
    def __init__(self, R: float, phi: float): ...
    @overload
    def __init__(self, Z: complex): ...
    def __init__(self, arg1, arg2=None):
        super().__init__()
        if type(arg1) is np.ndarray and arg2 is None:
            Z = np.mean(np.exp(1j*arg1))
            R = np.abs(Z)
            phi = np.angle(Z)

        elif type(arg1) is complex and arg2 is None:
            R = np.abs(Z)
            phi = np.angle(Z)

        elif type(arg1) is float and type(arg2) is float:
            R = arg1 
            phi = arg2 
        else:
            raise ValueError
        self.R = self._content['R'] = R
        self.phi = self._content['phi'] = phi








import numpy as np
import h5py



def get_r(f: h5py.File, cutoff_t):
    attrs = get_attrs(f)
    t = np.array(f['t'])
    theta = np.array(f['theta'])
    N = attrs['N']
    z = np.sum(np.exp(1j*theta), axis=1)/N
    r = np.abs(z)

    
    N_t = len(t)

    rs = 0
    N_cut = 0
    r2s = 0
    for i in range(N_t):
        if t[i] > cutoff_t:
            N_cut += 1
            rs += r[i]
            r2s += r[i]**2

    r_ave = rs / N_cut
    r2_ave = r2s / N_cut
    r_stdev = np.sqrt(r2_ave - r_ave**2)
    return rs / N_cut, r_stdev


def get_attrs(f):
    d = dict()
    for key in f.attrs:
        d[key] = f.attrs[key]
    return d



import time
import threading
import sys
import warnings
from pprint import pprint
import tqdm
from typing import List, Optional, overload

import numpy as np
from matplotlib import pyplot as plt
from scipy.fft import set_global_backend, fftn
import pyfftw
import h5py

from michael960lib.math import fourier
from michael960lib.common import overrides, experimental, deprecated
from michael960lib.common import ModifyingReadOnlyObjectError, IllegalActionError 
from torusgrid.grids import RealGrid2D, ComplexGrid2D, RealGrid1D, ComplexGrid1D, RealGridND, ComplexGridND
from torusgrid.dynamics import FancyEvolver, EvolverHistory, StateFunction

from ..core import KuramotoStateFunction, KuramotoBase
from ..storage import KuramotoWriter


class KuramotoEvolver(FancyEvolver):
    @overload
    def __init__(self, kura: KuramotoBase, dt: float): ...
    @overload
    def __init__(self, K: float, theta: np.ndarray, omega: np.ndarray, weight: np.ndarray, dt: float): ...
    def __init__(self, *args):
        if len(args) == 2:
            self.kura = args[0]
            self.dt = args[1]
        elif len(args) == 5:
            self.kura = KuramotoBase(args[0], args[1], args[2], args[3])
            self.dt = arg[4]
        else:
            raise ValueError('illegal arguments')

        theta = self.kura.theta.copy()

        self._thetagrid = RealGridND(self.kura.shape)
        self._thetagrid.set_psi(theta)
        self.theta = self._thetagrid.psi
        self.wk = fftn(self.kura.weight)
        self.omega = self.kura.omega.copy()

        super().__init__(self._thetagrid)

        self.Z = ComplexGridND(self.kura.shape)
        self.Z.initialize_fft()

        self.age = 0 
        self.theta_tmp = theta.copy()
        self.name = 'KuramotoND'
    
        self.K = self.kura.K
        self.info = {'system': self.name, 'dt': self.dt, 'K': self.K, 'shape': self.kura.shape,
            'dim': self.kura.rank,
            'omega_dist': self.kura.omega_params.get('type', '?'),
            'omega0': self.kura.omega_params.get('mean', np.nan),
            'gamma': self.kura.omega_params.get('scale', np.nan),
            'weight_type': self.kura.weight_params.get('type', '?'),

            'theta_label': self.kura.get_theta_label(),
            'omega_label': self.kura.get_omega_label(),
            'weight_label': self.kura.get_weight_label()
        }

        self.hdf = None
        self.is_recording = False

        self.set_display_format(
            '[{system}][dim={dim} shape={shape} K={K:5.2f}][weight: {weight_label}]' +\
            '[omega: {omega_label}]' +\
            '[dt={dt:6.3f} t={age:>09.3f}][R={R:4.2f} phi={phi:>5.2f}]'
        )

    def start_recording(self, fname):
        self.hdf = KuramotoWriter(fname, self.kura, self.age)
        self.hdf.file.attrs['dt'] = self.dt
        self.hdf.file.attrs['system'] = self.name
        self.is_recording = True

    def end_recording(self):
        self.hdf.close()

    def theta_dot(self):
        self.Z.psi[...] = np.exp(1j*self.theta_tmp)
        z = np.exp(1j*self.theta_tmp)
        self.Z.fft()
        self.Z.psi_k *= self.wk
        self.Z.ifft() 

        return self.K*np.imag(self.Z.psi/z) + self.omega

    @overrides(FancyEvolver)
    def step(self):
        self.theta_tmp[...] = self.theta
        k0 = self.theta_dot()
        self.theta_tmp[...] = self.theta + self.dt/2 * k0
        k1 = self.theta_dot()
        self.theta_tmp[...] = self.theta + self.dt/2 * k1
        k2 = self.theta_dot()
        self.theta_tmp[...] = self.theta + self.dt * k2
        k3 = self.theta_dot()

        self.theta += self.dt * (k0 + 2*k1 + 2*k2 + k3) / 6
        self.age += self.dt

    @overrides(FancyEvolver)
    def on_epoch_end(self, pb: tqdm.tqdm):
        super().on_epoch_end(pb)
        if self.is_recording:
            self.hdf.append(self.age, self.theta)

    @overrides(FancyEvolver)
    def on_nonstop_epoch_end(self):
        super().on_nonstop_epoch_end()
        if self.is_recording:
            self.hdf.append(self.age, self.theta)

    @overrides(FancyEvolver)
    def get_state_function(self) -> KuramotoStateFunction:
        return KuramotoStateFunction(self.theta)

    @overrides(FancyEvolver)
    def get_evolver_state(self) -> dict():
        _ = {
                'age': self.age
        }
        return _


# class KuramotoEvolver1D(KuramotoEvolver):
    # @overload
    # def __init__(self, kura: KuramotoBase, dt: float): ...
    # @overload
    # def __init__(self, K: float, theta: np.ndarray, omega: np.ndarray, weight: np.ndarray, dt: float): ...
    # def __init__(self, *args):
        # if len(args) == 2:
            # self.kura = args[0]
            # self.dt = args[1]
        # elif len(args) == 5:
            # self.kura = KuramotoBase(args[0], args[1], args[2], args[3])
            # self.dt = arg[4]
        # else:
            # raise ValueError('illegal arguments')

        # try:
            # assert len(self.kura.shape) == 1
        # except AssertionError:
            # raise ValueError('theta, omega, and weight must be one dimensional arrays of the same length')

        # theta = self.kura.theta.copy()
        # N = self.kura.shape[0]

        # self._thetagrid = RealGrid1D(N)
        # self._thetagrid.set_psi(theta)
        # self.theta = self._thetagrid.psi
        # self.wk = fft(self.kura.weight)
        # self.omega = self.kura.omega.copy()

        # super().__init__(self._thetagrid)

        # self.Z = ComplexGrid1D(N)
        # self.Z.initialize_fft()

        # self.age = 0 
        # self.theta_tmp = theta.copy()
        # self.name = 'Kuramoto1D'
    
        # self.K = self.kura.K
        # self.info = {'system': self.name, 'dt': self.dt, 'K': self.K, 'N': N,
            # 'dist': self.kura.omega_params.get('type', '?'),
            # 'omega0': self.kura.omega_params.get('mean', '?'),
            # 'weight': self.kura.weight_params.get('type', '?'),
            # 'nu': self.kura.weight_params.get('nu', '?')
        # }

        # self.hdf = None
        # self.is_recording = False

        # dt = self.dt
        # self.set_display_format(
            # '[{system}][N={N:<5} K={K:5.2f}][weight={weight} nu={nu:3.1f}] ' +\
            # '[dist={dist} omega0={omega0:<4.2f}]' +\
            # '[dt={dt:6.3f} t={age:>09.3f}][R={R:4.2f} phi={phi:>5.2f}]'
        # )

    # def start_recording(self, fname):
        # self.hdf = KuramotoHDF(fname, self.kura, self.age)
        # self.hdf.file.attrs['N'] = self.kura.shape[0]
        # self.hdf.file.attrs['dt'] = self.dt
        # self.hdf.file.attrs['system'] = self.name
        # self.is_recording = True

    # def end_recording(self):
        # self.hdf.close()

    # def theta_dot(self):
        # self.Z.psi[...] = np.exp(1j*self.theta_tmp)
        # z = np.exp(1j*self.theta_tmp)
        # self.Z.fft()
        # self.Z.psi_k *= self.wk
        # self.Z.ifft() 

        # return self.K*np.imag(self.Z.psi/z) + self.omega

    # @overrides(FancyEvolver)
    # def step(self):
        # self.theta_tmp[:] = self.theta
        # k0 = self.theta_dot()
        # self.theta_tmp[:] = self.theta + self.dt/2 * k0
        # k1 = self.theta_dot()
        # self.theta_tmp[:] = self.theta + self.dt/2 * k1
        # k2 = self.theta_dot()
        # self.theta_tmp[:] = self.theta + self.dt * k2
        # k3 = self.theta_dot()

        # self.theta += self.dt * (k0 + 2*k1 + 2*k2 + k3) / 6
        # self.age += self.dt

    # @overrides(FancyEvolver)
    # def on_epoch_end(self, pb: tqdm.tqdm):
        # super().on_epoch_end(pb)
        # if self.is_recording:
            # self.hdf.append(self.age, self.theta)

    # @overrides(FancyEvolver)
    # def on_nonstop_epoch_end(self):
        # super().on_nonstop_epoch_end()
        # if self.is_recording:
            # self.hdf.append(self.age, self.theta)

    # @overrides(FancyEvolver)
    # def get_state_function(self) -> KuramotoStateFunction:
        # return KuramotoStateFunction(self.theta)

    # @overrides(FancyEvolver)
    # def get_evolver_state(self) -> dict():
        # _ = {
                # 'age': self.age
        # }
        # return _






import time
import threading
import sys
import warnings
from pprint import pprint
import tqdm
from typing import List, Optional, overload, Tuple

import numpy as np
from matplotlib import pyplot as plt
from scipy.fft import fft2, ifft2, rfft2, irfft2, set_global_backend
import pyfftw

from michael960lib.math import fourier
from michael960lib.common import overrides, experimental, deprecated
from michael960lib.common import ModifyingReadOnlyObjectError, IllegalActionError

from torusgrid.grids import RealGrid2D, ComplexGrid2D, RealGrid1D, ComplexGrid1D, StateFunction

from ..math.math import torus_distance


class ParametrizedWeight:
    def __init__(self, shape: Tuple[int]):
        self._params = dict()
        self.name = 'NULL'
        self.label = 'NULL'
        self.shape = shape
        self.rank = len(shape)
        self._weight = np.zeros(shape)
        self._origin = tuple(0 for i in range(self.rank))


    
    def get_params(self):
        return self._params.copy()

    def get_weight(self):
        return self._weight.copy()

    def get_label(self):
        return self.label



class PowerLaw(ParametrizedWeight):
    def __init__(self, shape: Tuple[int], nu: float):
        super().__init__(shape)
        self._params['nu'] = nu
        self._params['type'] = self.name = 'power'
        self.label = f'power({nu})'

        for k, v in np.ndenumerate(self._weight):
            if k == self._origin:
                self._weight[k] = 0
            else:
                r = torus_distance(self.shape, k)
                self._weight[k] = 1/r**nu
        

class NearestNeighbor(ParametrizedWeight):
    def __init__(self, shape: Tuple[int]):
        super().__init__(shape)
        self._params['type'] = self.name = 'nearest-neighbor'
        for i in range(self.rank):
            p1 = tuple(1 if j==i else 0 for j in range(self.rank))
            p2 = tuple(-1 if j==i else 0 for j in range(self.rank))

            self._weight[p1] = 1
            self._weight[p2] = 1
        self.label = 'nearest'







import time
import threading
import sys
import warnings
from pprint import pprint
import tqdm
from typing import List, Optional
from datetime import datetime
import numpy as np
from matplotlib import pyplot as plt
from scipy.fft import fft2, ifft2, rfft2, irfft2, set_global_backend, fft, ifft
import pyfftw
import h5py

from michael960lib.math import fourier
from michael960lib.common import overrides, experimental, deprecated
from michael960lib.common import ModifyingReadOnlyObjectError, IllegalActionError 
from torusgrid.grids import RealGrid2D, ComplexGrid2D, RealGrid1D, ComplexGrid1D
from torusgrid.dynamics import FancyEvolver, EvolverHistory, StateFunction

from ..core import KuramotoStateFunction, KuramotoBase


class KuramotoWriter:
    def __init__(self, fname: str, kura: KuramotoBase, initial_age=0):
        self.fname = fname
        self.kura = kura
        self.file = h5py.File(fname, 'w')

        self.file.create_dataset('omega', dtype='f', data=kura.omega.tolist())
        for key in kura.omega_params:
            self.file['omega'].attrs[key] = kura.omega_params[key]

        self.file.create_dataset('weight', dtype='f', data=kura.weight.tolist())
        for key in kura.weight_params:
            self.file['weight'].attrs[key] = kura.weight_params[key]

        self.file.create_dataset('theta', dtype='f', data=[kura.theta.tolist()], 
                chunks=(1, *kura.shape), maxshape=(None, *kura.shape))
        for key in kura.theta_params:
            self.file['theta'].attrs[key] = kura.theta_params[key]


        self.file.create_dataset('shape', dtype='i8', data=list(kura.shape))
        self.file.attrs['K'] = kura.K
        self.time = datetime.now()
        self.file.attrs['time'] = self.time.strftime('%Y.%m.%d-%H:%m:%S')
        self.file.create_dataset('t', dtype='f', data=[initial_age], chunks=(1,), maxshape=(None,))
        self._chunk_ind = 0

    def set_datetime_format(self, fmt='%Y.%m.%d-%H:%m:%S'):
        self.file.attrs['time'] = self.time.strftime(fmt)

    def append(self, t: float, theta: np.ndarray):
        self._chunk_ind += 1
        self.file['theta'].resize(self._chunk_ind+1, axis=0)
        self.file['theta'][self._chunk_ind] = theta.tolist()
        self.file['t'].resize(self._chunk_ind+1, axis=0)
        self.file['t'][self._chunk_ind] = t

    def get_file(self):
        return self.file
    
    def check_well_formed(self): 
        r = True
        return r

    def close(self):
        self.file.close()

class KuramotoReader:
    def __init__(self, fname: str):
        self.fname = fname
        self.file = h5py.File(fname, 'r')

        self.K = self.file.attrs['K']
        self.shape = np.array(self.file['shape'])

        self.theta_params = dict(self.file['theta'].attrs)
        self.theta = self.file['theta']
        self.theta_seed = self.theta_params.get('seed', '?')

        self.omega_params = dict(self.file['omega'].attrs)
        self.omega = np.array(self.file['omega'])
        self.omega_dist = self.omega_params.get('type', '?')
        self.omega0 = self.omega_params.get('mean', np.NAN)
        self.gamma = self.omega_params.get('scale', np.NaN)
        self.omega_seed = self.omega_params.get('seed', '?')

        self.weight_params = dict(self.file['weight'].attrs)
        self.weight = np.array(self.file['weight'])
        self.nu = self.weight_params.get('nu', np.NaN)
        self.weight_type = self.weight_params.get('type', '?')

        self.t = np.array(self.file['t'])
        self.Nt = len(self.t)
        self.t_max = self.t[-1]

        self.dt = self.file.attrs['dt']

        self.datetime = self.file.attrs.get('time', '?')

    def get_dimensions(self):
        return len(self.shape)






import numpy as np
from matplotlib import pyplot as plt
from pprint import pprint
from scipy.fft import fft2, ifft2
import sys
import matplotlib
import h5py
from datetime import datetime


class Kuramoto2d:
    def __init__(self, Nx, Ny, K):
        self.Nx = Nx
        self.Ny = Ny
        self.theta = np.zeros((Nx, Ny))
        self.omega = np.zeros((Nx, Ny))
        self.w = np.zeros((Nx, Ny))
        self.K = K

        self.meshX, self.meshY = np.meshgrid(np.arange(0, Nx), np.arange(0, Ny))

    def init_freq(self, omega0, gamma, seed=None):

        self.omega0 = omega0
        self.gamma = gamma

        if not seed is None:
            np.random.seed(seed)
        self.omega = np.random.standard_cauchy((self.Nx, self.Ny)) * gamma + omega0

    def set_freq(self, omega):
        self.omega = omega

    def init_phase(self, seed=None):
        if not seed is None:
            np.random.seed(seed)
        self.theta = np.random.uniform(0, np.pi*2, (self.Nx, self.Ny))

    def set_phase(self, theta):
        self.theta = theta

    def set_weights(self, w):
        self.w = w
        self.wk = fft2(self.w)
        self.Zmax = np.sum(w)
    
    def theta_dot(self, theta1):
        z = np.exp(1j*theta1)
        zk = fft2(z)
        Z = ifft2(self.wk*zk)
        return self.omega + self.K*np.abs(Z)*np.sin(np.angle(Z)-np.angle(z))

    def evolve(self, dt, t_max, t_digits=3, cycle=29, extra_prefix=''):
        t = 0
        d = 0
        while t < t_max:
            k1 = self.theta_dot(self.theta)
            k2 = self.theta_dot(self.theta + dt*k1/2)
            k3 = self.theta_dot(self.theta + dt*k2/2)
            k4 = self.theta_dot(self.theta + dt*k3)
            self.theta += dt * (k1 + 2*k2 + 2*k3 + k4) / 6

            t += dt
            d += 1

            if d >= cycle:
                d = 0
                Z0 = self.get_Z0()
                R0 = np.round(np.abs(Z0), 3)
                Phi0 = np.round(np.angle(Z0), 3)
                sys.stdout.write(f'\r  {extra_prefix}   {np.round(t, t_digits)}/{t_max}   |   Z0={R0, Phi0}                                  ')
                sys.stdout.flush()
        return t

    def evolve_and_record_hdf(self, dt, batch_size, t_max, fname):
        f = h5py.File(fname, 'w', )
        f.create_dataset('omega', dtype='f', data=self.omega.tolist())
        f.create_dataset('weight', dtype='f', data=self.w.tolist())
        f.create_dataset('theta', dtype='f', data=[self.theta.tolist()], chunks=(1, self.Nx, self.Ny), maxshape=(None, self.Nx,
                                                                                                                 self.Ny))
        t = 0
        f.create_dataset('t', dtype='f', data=[t], chunks=(1,), maxshape=(None,))
        f.attrs['time'] = datetime.now().strftime('%Y.%m.%d-%H:%m:%S') 

        chunk_ind = 0

        while t < t_max:
            chunk_ind += 1
            delta_t = self.evolve(dt, dt*batch_size, cycle=batch_size//2, extra_prefix=f'{np.round(t, 3)}/{t_max}\t  |   ')
            t += delta_t

            f['theta'].resize(chunk_ind+1, axis=0)
            f['theta'][chunk_ind] = self.theta.tolist()
            f['t'].resize(chunk_ind+1, axis=0)
            f['t'][chunk_ind] = t
        sys.stdout.flush()
        print()

        return f

    def get_Z0(self):
        return get_Z0(self.theta, self.Nx, self.Ny)

    def get_Z(self):
        return get_Z(self.theta, self.wk)

    def draw_plots(self, show=True):
        ax1 = plt.subplot(1, 3, 1)
        ax2 = plt.subplot(1, 3, 2)
        ax3 = plt.subplot(1, 3, 3)

        cf1 = ax1.pcolormesh(self.meshX, self.meshY, self.theta % (2*np.pi), cmap='hsv', vmin=0, vmax=np.pi*2, shading='auto')
        ax1.set_aspect('equal', 'box')
        ax1.set_xticks([0, self.Nx])
        ax1.set_yticks([0, self.Ny])

        Z = self.get_Z() 
        
        cf2 = ax2.imshow(complex_array_to_rgb(Z/self.Zmax, rmax=1), extent=(0,self.Nx,0,self.Ny), origin='lower',
                         aspect='equal')

        ax2.set_aspect('equal', 'box')
        ax2.set_xticks([0, self.Nx])
        ax2.set_yticks([0, self.Ny])
        
        Z0 = self.get_Z0()

        ax3.set_xlim(-1, 1)
        ax3.set_ylim(-1, 1)
        ax3.plot([np.real(Z0)], [np.imag(Z0)])
        ax3.set_aspect('equal', 'box')

        ax3.plot([0, 0], [-1, 1], lw=1, color='black')
        ax3.plot([-1, 1], [0, 0], lw=1, color='black')
        
        plt.colorbar(cf1, ax=[ax1, ax2, ax3], location='top', fraction=0.05)

        if show:
            plt.show()
        
        return ax1, ax2, ax3

def get_Z0(theta, Nx, Ny):
    return np.sum(np.exp(1j*theta)) / (Nx*Ny)

def get_Z(theta, wk):
    zk = fft2(np.exp(1j*theta))
    Z = ifft2(wk*zk)
    return Z


def power_weights(nu, Nx, Ny):
    w = np.zeros((Nx, Ny))

    for i in range(Nx):
        for j in range(Ny):
            if i == 0 and j == 0:
                w[i][j] = 0
            else:
                w[i][j] = 1 / np.sqrt(min(i**2, (i-Nx)**2) + min(j**2, (j-Ny)**2)) ** nu
    return w


#--- plot utils
def complex_array_to_rgb(X, theme='dark', rmax=None):
    '''Takes an array of complex number and converts it to an array of [r, g, b],
    where phase gives hue and saturaton/value are given by the absolute value.
    Especially for use with imshow for complex plots.'''
    absmax = rmax or np.abs(X).max()
    Y = np.zeros(X.shape + (3,), dtype='float')
    Y[..., 0] = (np.angle(X)) / (2 * np.pi) % 1
    if theme == 'light':
        Y[..., 1] = np.clip(np.abs(X) / absmax, 0, 1)
        Y[..., 2] = 1
    elif theme == 'dark':
        Y[..., 1] = 1
        Y[..., 2] = np.clip(np.abs(X) / absmax, 0, 1)
    Y = matplotlib.colors.hsv_to_rgb(Y)
    return Y






import numpy as np


def torus_distance(shape, point):
    rank = len(shape) 
    r2 = 0
    for i in range(rank):
        ri2 = min(point[i], shape[i]-point[i])**2
        r2 += ri2
    return np.sqrt(r2)




import os
import h5py

def get_filename(dirct, prefix, suffix, form):
    i = 0
    files = os.listdir(dirct)
    while True:
        fname = prefix + form.format(i) + suffix
        if not (fname in files):
            return fname
        i += 1

def get_hdf(dirct, mask_func):
    files = os.listdir(dirct)
    files_hdf = []
    for f in files:
        if f.endswith('.hdf5'):
            files_hdf.append(f)

    opened_hdf = [] 
    for fh in files_hdf:
        hdf_file = h5py.File(f'{dirct}/{fh}', 'r')
        if mask_func(hdf_file):
            opened_hdf.append(hdf_file)
        else:
            hdf_file.close()
    return opened_hdf
